#!/usr/bin/env python3
import asyncio
import websockets
import sys

messages = asyncio.Queue()
cmds = asyncio.Queue()
cp = ''


async def send_loop(websocket):
    while True:
        if not cmds.empty():
            cmd = cmds.get_nowait()
            await websocket.send(cmd)
        if not messages.empty():
            reply = messages.get_nowait()
            print(reply.strip())
        await asyncio.sleep(0)


def got_stdin_data():
    msg = sys.stdin.readline()
    cmds.put_nowait(msg)


async def receive_loop():
    print("one")
    if len(sys.argv)>=2:
        print(sys.argv[1])
        server = sys.argv[1]
    else:
        server = input('connect to:')
    async with websockets.connect('ws://' + server + ':5005') as websocket:
        task = asyncio.Task(send_loop(websocket))

        while True:
            reply = await websocket.recv()
            messages.put_nowait(reply)
try:
    asyncio.get_event_loop().add_reader( sys.stdin, got_stdin_data)
    asyncio.get_event_loop().run_until_complete(receive_loop())
except KeyboardInterrupt:
    pass


from nodewire.NodeWire import NodeWire

class Port:
    def __init__(self, name, props='Digital OUT', writing=None, reading=None):
        self.store = 0
        self.name = name
        self.props = props
        self.writing = writing
        self.reading = reading

    def read(self):
        if(self.reading): self.reading()
        return  self.store

    def write(self, val):
        if(self.writing): self.writing(val)
        self.store = val

    def properties(self):
        return self.props

class Node:
    def __init__(self, nodename='node01', nw=None):
        self.nw = NodeWire(nodename, process=self.process) if nw is None else nw
        self.nw.debug = True
        self.ports = []

    def process(self, msg):
        if msg.Command == 'get' or msg.Command == 'getportvalue':
            if msg.Params[0] == 'properties':
                p = [pp for pp in self.ports if pp.name == msg.Params[1]]
                if len(p) != 0: self.nw.send(msg.Sender,'properties', p[0].name, p[0].props)
            elif msg.Params[0] == 'ports':
                ps = ''
                for p in self.ports: ps = ps + ' ' + p.name
                self.nw.send(msg.Sender, 'ports', ps)
            else:
                v = self.get(msg.Params[0])
                self.nw.send(msg.Sender, 'portvalue', msg.Params[0], str(v))
        elif msg.Command == 'set' or msg.Command == 'setportvalue':
            self.set(msg.Params[0], msg.Params[1])
            self.nw.send(msg.Sender, 'portvalue', msg.Params[0], self.get(msg.Params[0]))

    def get(self, port):
        p = [pp for pp in self.ports if pp.name==port] # filter(lambda pp: pp.name==port, self.ports)
        if len(p) != 0: return p[0].read()
        else: return 'error'

    def set(self, port, val):
        p = [pp for pp in self.ports if pp.name == port]
        if len(p) != 0: p[0].write(val)

if __name__ == '__main__':
    node = Node()
    node.ports.append(Port('Power'))
    node.ports.append(Port('LED', 'Digital IN'))

    node.nw.run()


'''
Copyright (c) 2016, nodewire.org
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
   This product includes software developed by nodewire.org.
4. Neither the name of nodewire.org nor the
   names of its contributors may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY nodewire.org ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL nodewire.org BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
'''
import time, threading
import requests
from nodewire.splitter import split, getsender, getnode
import asyncio
from nodewire.client import client
import json
import uuid
import inspect

debug = False # False for deployment

class Message:
    def __init__(self, msg):
        words = split(msg)
        self.Address = words[0]
        self.Command = words[1] if len(words)>1 else words[0]
        self.Params = words[2:-1] if len(words)>2 else words[0]
        self.Port = words[2] if len(words)>=4 else None
        try:
            self.Value = json.loads(words[3]) if len(words)>=5 else None
        except:
            self.Value = None
        self.Sender = words[-1]

    def __str__(self):
        return self.Address + ' ' + self.Command + ' ' + ' '.join(p for p  in self.Params) + ' ' + self.Sender


class NodeWire:
    def __init__(self, node_name='node', server='cloud1.nodewire.org', process=None):
        self.name = node_name
        self.type = node_name
        self.server_address = server
        self.gateway = ''
        self.id = 'None'
        self.callbacks = {}
        self.terminated = False
        self.client = client()
        self.called_connected =  False
        self.connected = False

        self.ack = False
        self.waiting_config = False

        try:
            self.readconfig()
            print(self.uuid)
        except Exception as ex:
            print('Failed to read configuration file. Creating new config...')
            self.uuid = str(uuid.uuid4())
            print(f'New UUID is {self.uuid}')
            self.token = 'None'
            self.name = node_name
            self.id = 'None'
            self.saveconfig()

        if self.process_command:
            self.client.received = self.process_command

        self.process = process
        self.on_connected = None
        self.debug = False

    def saveconfig(self):
        file = open('nw.cfg', 'w')
        file.write(json.dumps({
            'uuid': self.uuid,
            'token': self.token,
            'name': self.name,
            'id': self.id,
            'gateway': self.gateway
        }))
        file.close()

    def readconfig(self):
        file = open('nw.cfg', 'r')
        config = json.loads(file.read())
        self.uuid = config['uuid']
        self.token = config['token']
        self.name = config['name']
        self.id = config['id']
        self.gateway = config['gateway']
        file.close()

    async def start(self, loop):
        if self.token == 'None':
            await self.client.sendasync('cp Gateway id={}\n'.format(self.uuid))
            self.waiting_config = True
        else:
            await self.client.sendasync('cp Gateway key={} {}\n'.format(self.token, self.uuid))
        self.connected = True

    def send(self, Node, Command, *Params):
        if self.connected:
            try:
                #todo rewrite this as format function
                cmd = Node + ' ' + Command + ' ' + ' '.join(param for param in Params) + (' ' + self.name if len(Params) != 0 else self.name)
                if self.debug:print(cmd)
                self.client.send(cmd+'\n')
                return True
            except Exception as ex:
                if self.debug:print('failed to send command over LAN')
                self.connected = False
                return False

    async def pinger(self):
        if self.debug: print('pinging...')
        while not self.ack:
            self.send('cp', 'ThisIs', self.id)
            await asyncio.sleep(5)

    async def keepalive(self):
        await asyncio.sleep(60)
        while True:
            self.ack = False
            try:
                self.send('cp', 'keepalive')
            except:
                await self.start(asyncio.get_event_loop())
            await asyncio.sleep(5)
            if not self.ack:
                if self.debug: print('didn\'t recieve ack')
                self.client.close_connection()
                self.connected = False
                await self.start(asyncio.get_event_loop())
            await asyncio.sleep(300)

    def when(self, cmd, func):
        self.callbacks[cmd] = func

    async def process_command(self, cmd):
        self.last = time.time()
        if cmd == 'disconnected':
            self.connected = False
            return
        msg = Message(cmd)

        if self.debug: print(cmd)

        if msg.Command == 'ack':
            self.ack = True
        elif msg.Command == 'gack' and not self.called_connected:
            print('connected')
            if self.waiting_config:
                self.waiting_config = False
                self.gateway = msg.Address.split(':')[0]
                self.token = msg.Address.split(':')[1]
                self.saveconfig()
                self.client.close_connection()
                self.connected = False
                await self.start(asyncio.get_event_loop())
            if self.on_connected:
                self.called_connected = True
                if inspect.iscoroutinefunction(self.on_connected):
                    asyncio.create_task(self.on_connected())
                else:
                    self.on_connected()
        elif msg.Command == 'authfail':
            if self.token != 'None':
                if self.debug: print('we have been deleted')
                self.token = 'None'
                self.saveconfig()
        elif msg.Command == 'ping':
            self.ack = False
            asyncio.Task(self.pinger())
        elif msg.Command == 'get' and msg.Params[0] == 'id':
            self.send(msg.Sender, 'id', self.id)
        elif msg.Command == 'get' and msg.Params[0] == 'type':
            self.send(msg.Sender, 'type', self.type)
        elif msg.Command == 'set' and msg.Params[0] == 'id':
            self.id = msg.Params[1]
            self.saveconfig()
        elif msg.Command == 'set' and msg.Params[0] == 'name':
            self.name = msg.Params[1]
            self.saveconfig()
            self.send(msg.Sender, 'ThisIs')
        else:
            if self.process:
                self.process(msg)
        if msg.Command == 'portvalue':
            signal = (msg.Sender.split(':')[1] if ':' in msg.Sender else msg.Sender) + '.' + msg.Params[0]
            if signal in self.callbacks:
                self.callbacks[signal](msg)
        elif msg.Command in self.callbacks:
                self.callbacks[msg.Command](msg)

    async def run_async(self):
        loop = asyncio.get_event_loop()
        await asyncio.gather(
            asyncio.create_task(self.client.connect(serverHost=self.server_address)),
            asyncio.create_task(self.start(loop)),
            asyncio.create_task(self.keepalive())
        )

    def run(self, tsk=None):
        loop = asyncio.get_event_loop()
        tasks = [
            loop.create_task(self.client.connect(serverHost=self.server_address)),
            loop.create_task(self.start(loop)),
            loop.create_task(self.keepalive())
        ]
        if tsk:
            if self.debug: print('starting task')
            tasks+=[loop.create_task(tsk)]
        wait_tasks = asyncio.wait(tasks)
        try:
            loop.run_until_complete(wait_tasks)
        except KeyboardInterrupt:
            # Canceling pending tasks and stopping the loop
            asyncio.gather(*asyncio.Task.all_tasks()).cancel()
            # Stopping the loop
            loop.stop()
            # Received Ctrl+C
            loop.close()

if __name__ == '__main__':
    nw = NodeWire('pyNode')
    nw.debug = True
    nw.run()




import asyncio
import json

class db(object):
    def __init__(self, nw):
        self.nw = nw

    async def get(self, table, query, projection=None, options=None):
        """ahmad = await db.get('test', {'age', 20})"""
        queue = asyncio.Queue()
        if projection is None:
            projection = {}
        if options is None:
            options = {}

        def db_result(msg):
            queue.put_nowait(msg.Value)

        if projection == {} and options == {}:
            self.nw.send('db','get', table, json.dumps(query))
        else:
            self.nw.send('db','get', table, json.dumps(query), json.dumps(projection), json.dumps(options))
        self.nw.when('db.'+ table, db_result)
        return await queue.get()

    async def first(self, table, query, projection=None, options=None):
        result = await self.get(table, query, projection, options)
        if result:
            return result[0]
        else:
            return None

    async def last(self, table, query, projection=None, options=None):
        result = await self.get(table, query, projection, options)
        if result:
            return result[-1]
        else:
            return None

    def index(self, table, keys, options = None):
        if options:
            self.nw.send('db', 'set', table, 'index', keys, options)
        else:
            self.nw.send('db', 'set', table, 'index', keys)

    def drop(self, table):
        """db.drop('test')"""
        self.nw.send('db','set', table, 'drop')

    def remove(self, table, query):
        """db.remove('test', {'age',40})"""
        self.nw.send('db', 'set', table, 'remove', json.dumps(query))

    def set(self, table, value):
        """db.set('test', {'name':'ahmad', 'age': 40})"""
        return self.nw.send('db', 'set', table, json.dumps(value))

    async def set_wait(self, table, value):
        """db.set('test', {'name':'ahmad', 'age': 40})"""
        queue = asyncio.Queue()

        def db_result(msg):
            queue.put_nowait(msg.Value)

        self.nw.send('db', 'set', table, json.dumps(value))
        self.nw.when('db.' + table, db_result)
        return await queue.get()

    async def aggregate(self, table, query):
        """oldest = await db.aggregate('test', [{"$group" : {"_id" : 1, "oldest" : {"$max" : "$age"}}}])"""
        queue = asyncio.Queue()

        def db_result(msg):
            queue.put_nowait(msg.Value)

        self.nw.send('db', 'get', table, 'aggregate', json.dumps(query))
        self.nw.when('db.' + table, db_result)
        return await queue.get()


if __name__ == '__main__':
    from nodewire.control import control
    async def myloop():
        await asyncio.sleep(10)
        mydb = db(ctrl.nw)
        ahmad = await mydb.get('test', {})
        print(ahmad)

    def connected():
        mydb = db(ctrl.nw)
        mydb.set('test', {'name':'ahmad sadiq', 'age': 43})

    ctrl = control()
    ctrl.nw.on_connected = connected
    ctrl.nw.run(myloop())


#!/usr/bin/env python
"""
Client side: use sockets to send data to the server, and print server's
reply to each message line; 'localhost' means that the server is running
on the same machine as the client, which lets us test client and server
on one machine; to test over the Internet, run a server on a remote
machine, and set serverHost or argv[1] to machine's domain name or IP addr;
Python sockets are a portable BSD socket interface, with object methods
for the standard socket calls available in the system's C library;
"""

import sys
from socket import * # portable socket interface plus constants
from datetime import *
import time
#global lastSN

MAX = 128
class TCPClient():
    def __init__(self,serverHost='138.197.6.173',serverPort=10001,clientID='0000'):
        self.serverHost = serverHost # server name
        self.serverPort = serverPort # non-reserved port used by the server
        self.socketobj = socket(AF_INET, SOCK_STREAM)  # make a TCP/IP socket object
        self.received = None

    def connect(self):
        #sockobj = socket(AF_INET, SOCK_STREAM)  # make a TCP/IP socket object
        self.socketobj.connect((self.serverHost, self.serverPort))  # connect to server machine + port

    def send(self,data,size=MAX):#command should by byte string
        self.socketobj.sendall(data.encode())

    def receive(self, size=MAX):
        #todo wait for new line
        while True:
            data = ''
            try:
                data += self.socketobj.recv(MAX).decode()
                while data:
                    if '\n' in data:
                        lines = data.splitlines()
                        if len(lines) == 1:
                            self.received(lines[0])
                            data = ''
                        else:
                            for line in lines[:-1]:
                                if len(line.strip()) != 0:
                                    self.received(line)
                            data = lines[-1]
                    data += self.socketobj.recv(MAX).decode()
                if data == '':
                    self.received('disconnected')
                    return
            except Exception as ex:
                print('tcp error:' + str(ex))
                try:
                    #self.connect()
                    self.received('disconnected')
                except Exception as ex1:
                    print('another error' + str(ex1))
                    self.received('disconnected')
                return

if __name__ == '__main__':
    pass



#!/usr/bin/env python3
from nodewire.NodeWire import NodeWire
import asyncio
import configparser
from nodewire.client import client
import sys
import json
import time

class Node:
    def __init__(self, client):
        self.nw = NodeWire(client, process=self.process)
        self.terminated = False

    def prompt(self):
        print('---- '),

    def process(self, msg):
        if msg.Command == 'portvalue':
            if msg.Params[0] == 'script':
                s = json.loads(msg.Params[1])
                for line in s:
                    print(line)
                self.prompt()

    def got_stdin_data(self):
        cmd = sys.stdin.readline()
        if cmd.startswith('quit'):
            self.terminated = True
            print('closing...')
            quit()
        else:
            node.nw.send('re', 'set', 'scriptlet', '"' + cmd.strip() + '"')


if __name__ == '__main__':
    global Prompt
    while True:
        print('NodeWire Interactive shell')
        print('Copyright 2017')
        print('connecting ...')
        config = configparser.RawConfigParser()

        node = Node(client())
        loop = asyncio.get_event_loop()
        loop.add_reader(sys.stdin, node.got_stdin_data)
        node.prompt()

        try:
            node.nw.run()
        except Exception  as ex:
            print(ex)

        time.sleep(10)



def opposite(c):
    if c== '{': return '}'
    if c== '[': return ']'
    if c== '(': return ')'
    return  c

def getsender(s):
    return s[s.rfind(' '):].strip()

def getnode(s):
    return s[:s.find(' ')].strip()

def getinstance(s):
    node = getnode(s)
    if ":" in node:
        return node.split(':')[0].strip()
    else:
        sender = getsender(s)
        if ':' in sender:
            return sender.split(':')[0].strip()
        else:
            return None

def getnodeonly(s):
    nn = getnode(s)
    if ':' in nn:
        return nn.split(':')[1].strip()
    else:
        return nn

def getsenderonly(s):
    nn = getsender(s)
    if ':' in nn:
        return nn.split(':')[1].strip()
    else:
        return nn


def getfulladdress(s):
    g = getinstance(s)
    n = getnodeonly(s)
    return g + ':' + n

def getfullsender(s):
    n = getsender(s)
    if ':' in n: return n
    g = getinstance(s)
    #n = getsenderonly(s)
    return g + ':' + n

def split(s):
   s = s.strip()
   tokens = []
   token = ''
   tokcount = 0
   sep = ''
   isparam = False
   for c in s:
      if c in [' ', '[', '(', '{', '"', "'"] and sep == '':
          if len(token) != 0 and c == ' ':
              if token == '=' or isparam:
                  isparam = not isparam
                  tokens[-1]+= token
              else:
                tokens.append(token)
              token = ''
              sep = ''
              tokcount=0
          elif c!= ' ':
              sep = c
              tokcount+=1
      elif c in [' ', '[', '(', '{'] and c==sep:
          tokcount+=1
          token += c
      elif c in [']', ')', '}', '"'] and c == opposite(sep):
          tokcount-=1
          if tokcount == 0:
              if token == '=' or isparam:
                  isparam = not isparam
                  tokens[-1]+= token
              else:
                  tokens.append( sep + token + c)
              token = ''
              sep = ''
          else:
              token += c
      else:
          token += c
   if token!='': tokens.append(token)
   return tokens

if __name__ == '__main__':
    cmd = 'one to { "name" : "ahmad sadiq", children [ "a", "b", {"name": "c"}] } "Ahmad Sadiq" { "name" : "ahmad sadiq", children [ "a", "b", {"name": "c"}] }'
    cmd2 = 'one two "Ahmasd Sadiq" three "Garba Shehu"'
    cmd3 = "sadhhjf:cp getnodes checksum = 201 otp = ewhjkfewjknfejknfewkj ah"
    cmd4 = 'script send ["var=1", "when node[0]==1 do var=2"]'
    print (split(cmd))
    print (split(cmd2))
    print (split(cmd3))
    print (split(cmd4))



#!/usr/bin/env python
import asyncio
import time
import sys
import glob
from serial import Serial
from nodewire.client import client
from nodewire.splitter import split, getnode, getsender, getinstance, getfulladdress, getsenderonly, getnodeonly, getfullsender
import requests
import configparser

instancename = ""
cp = ""
user = None
password = None
baudrate = 38400
server = 'http://cloud1.nodewire.org'

serialport = None

class ucolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

class wcolors:
    HEADER = ''
    OKBLUE = ''
    OKGREEN = ''
    WARNING = ''
    FAIL = ''
    ENDC = ''
    BOLD = ''
    UNDERLINE = ''

if sys.platform.startswith('win'):
    bcolors = wcolors
else:
    bcolors = ucolors

def start():
    global instancename, cp, user, password, server
    config = configparser.RawConfigParser()
    try:
        config.read('nw.cfg')
        if 'server' in config: server = 'http://' + config['server']['address']
    except:
        print('could not find cp.cfg')
        print('please run from a directory that contains this file')
        quit()

    user = str(config.get('user', 'account_name'))
    password = str(config.get('user', 'password'))
    while True:
        with requests.Session() as s:
            try:
                res = s.post(server + '/login', data={'email': user, 'pw': password})
                if res.ok:
                    r = s.get(server + '/config').json()
                    instancename = str(r['instance'])
                    cp = r['server']
                    break
                else:
                    print(bcolors.FAIL + 'authentication failure' + bcolors.ENDC)
                    quit()
            except Exception as ex:
                print(
                    bcolors.WARNING + 'could not connect to cloud service. will retry after 30 seconds.' + bcolors.ENDC)
                time.sleep(30)

def openSerial():
    global serialport
    tries = 10
    print('searching for serial port...')

    while True:
        if sys.platform.startswith('win'):
            ports = ['COM%s' % (i + 1) for i in range(256)]
        elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
            # this excludes your current terminal "/dev/tty"
            ports = glob.glob('/dev/tty[A-Za-z]*')
        elif sys.platform.startswith('darwin'):
            ports = glob.glob('/dev/tty.*')
        else:
            raise EnvironmentError('Unsupported platform')
        sys.stdout.write(".")
        sys.stdout.flush()
        for port in ports:
            try:
                serialport = Serial(port, baudrate, timeout=1, write_timeout=1)
                serialport.reset_input_buffer()
                serialport.write('any ping cp\n'.encode())
                time.sleep(2)
                msg = serialport.readline().decode()
                serialport.write('any ping cp\n'.encode())
                msg = serialport.readline().decode()
                if msg.startswith('cp ThisIs') or msg.startswith('re ThisIs'):
                    print(bcolors.BOLD + 'serialport is on ' + port + bcolors.ENDC)
                    return serialport
                serialport.close()
            except Exception as exc:
                pass
        tries -= 1
        if tries == 0:
            print(bcolors.FAIL + 'giving up on serial port. will not support serial devices in this run' + bcolors.ENDC)
            return None
        time.sleep(30)


class Gateway():
    def __init__(self, client):
        self.scmd = bytearray()
        self.whenLastReceived = time.time()
        self.client = client
        self.sendqueue = asyncio.Queue()
        self.connected = False
        self.terminate = False
        self.ack = False

    async def keepalive(self):
        await asyncio.sleep(60)
        while True:
            self.ack = False
            try:
                if self.connected:
                    print('keepalive')
                    self.client.send('cp keepalive ' + user+'\n')
                else:
                    await self.start(asyncio.get_event_loop())
            except:
                await self.start(asyncio.get_event_loop())
            await asyncio.sleep(10)
            if not self.ack:
                print('didn\'t recieve ack')
                self.client.close_connection()
                await self.start(asyncio.get_event_loop())
            await asyncio.sleep(200)

    async def get_byte_async(self):
        while True:
            c = s.read(1)
            if c == b'' :
                await asyncio.sleep(0)
            else:
                s.write(c)
                self.whenLastReceived = time.time()
                return c

    async def uart_receive(self):
        while not self.terminate:
            b = await self.get_byte_async()
            if b!= b'\r' and b!=b'\n':
                self.scmd+=b
            elif len(self.scmd) != 0:
                print(bcolors.HEADER + self.scmd.decode('ascii') + bcolors.ENDC)
                if self.connected:
                    self.client.send(self.scmd.decode('ascii')+'\n')
                del self.scmd[:]


    async def start(self, loop):
        await self.client.sendasync('cp Gateway user={} pwd={} {}\n'.format(user,password,instancename))
        self.connected = True
        print('connected')
        serialport.write('any ping cp\n'.encode())

    def tcp_received(self, data):
        if data == 'disconnected':
            print(bcolors.WARNING + 'Lost Connection to Gateway' + bcolors.ENDC)
            self.connected = False
            # time.sleep(10)
            # tcp()
        else:
            self.ack = True
            datas = data.splitlines()
            for data in datas:
                node = getnode(data)
                if ':' in node:
                    nodeonly = node.split(':')[1]
                    data = data.replace(node, nodeonly).strip()
                self.sendqueue.put_nowait(data)

    async def uart_send(self):
        while not self.terminate:
            response = await self.sendqueue.get()
            #while time.time() - self.whenLastReceived < 0.01: await asyncio.sleep(0.01)
            s.write((response + '\n').encode())
            print(bcolors.OKBLUE + response + bcolors.ENDC)
            self.sendqueue.task_done()
            await asyncio.sleep(0)

if __name__ == '__main__':
    print(bcolors.BOLD + 'NodeWire Gateway' + bcolors.ENDC)
    print('version 1.0')
    print('Copyright 2020 nodewire.org')
    print('starting...')
    while True:
        start()
        s = openSerial()
        c = client()
        g = Gateway(c)
        c.received = g.tcp_received

        loop = asyncio.get_event_loop()

        tasks = [
            loop.create_task(c.connect(cp)),
            loop.create_task(g.start(loop)),
            loop.create_task(g.uart_receive()),
            loop.create_task(g.uart_send()),
            loop.create_task(g.keepalive())
        ]

        wait_tasks = asyncio.wait(tasks)
        try:
            loop.run_until_complete(wait_tasks)
        except KeyboardInterrupt:
            g.terminate = True


import asyncio
import inspect

class client():
    def __init__(self):
        self.reader = None
        self.writer = None
        self.received = None
        self.managed = True

    async def connect(self, serverHost='cloud1.nodewire.org', failed=None):
        while True:
            try:
                print('connecting to {} ...'.format(serverHost))
                self.reader, self.writer = await asyncio.open_connection(serverHost, 10001)
                while True:
                    try:
                        data = await self.reader.readline()
                        if len(data)==0 and self.received:
                            if inspect.iscoroutinefunction(self.received):
                                asyncio.create_task(self.received('disconnected'))
                            else:
                                self.received('disconnected')
                            break
                        if self.received:
                            try:
                                if inspect.iscoroutinefunction(self.received):
                                    asyncio.create_task(self.received(data.decode().strip()))
                                else:
                                    self.received(data.decode().strip())
                            except Exception as ex1:
                                print(str(ex1))
                                if not self.managed: raise
                    except ConnectionError as ex: # todo use specified exception type
                        print(str(ex))
                        if self.received:
                            self.received('disconnected')
                        break
                print('Close the socket')
                if failed!=None:
                    failed(serverHost)
                self.writer.close()
                self.writer = None
                await asyncio.sleep(10)
                print('trying to reconnect...')
            except Exception as Ex: # todo get the relevant exceptions: TimeoutError
                print('failed to connect')
                if failed!=None:
                    failed(serverHost)
                    break
                await asyncio.sleep(30)
                print('trying to reconnect...')

    def close_connection(self):
        if not self.writer is None:
            self.writer.close()
            self.writer = None

    async def sendasync(self, message):
        while(self.writer==None): await asyncio.sleep(1)
        self.writer.write(message.encode())

    def send(self, message):
        self.writer.write(message.encode())

if __name__ == '__main__':
    c = client()
    loop = asyncio.get_event_loop()
    try:
        loop.run_until_complete(c.connect())
    except KeyboardInterrupt:
        pass
    loop.close()


from setuptools import setup
# https://python-packaging.readthedocs.io/en/latest/minimal.html
# python setup.py sdist upload


def readme():
    with open('README.rst') as f:
        return f.read()


setup(name='nodewire',
      version='1.0.1',
      description='NodeWire',
      long_description=readme(),
      url='http://www.nodewire.org',
      author='Ahmad Sadiq',
      author_email='sadiq.a.ahmad@gmail.com',
      license='BSD',
      packages=['nodewire'],
      scripts=['bin/nw_script.py', 'bin/nw_client.py'],
      install_requires=[
            #'paho-mqtt',   # no longer needed
            'configparser',
            #'netifaces', # no longer needed
            'pyserial',
            'requests',
            'asyncio'
      ],
      zip_safe=False)


from nodewire.NodeWire import NodeWire
import json
from optparse import OptionParser
import inspect
import asyncio
import time

will_get = {}

class Node():
    def __init__(self, nw, name, gateway):
        self.name = name
        self.nw = nw
        self.gateway = gateway
        self.ports = {}

    def __iter__(self):
        self.iterobj = iter(self.ports)
        return self.iterobj

    def __next__(self):
        next(self.iterobj)

    def __unicode__(self):
        return self.name + str(self.ports)

    def __repr__(self):
        return self.name + str(self.ports)

    def __str__(self):
        return self.name + str(self.ports)

    def __contains__(self, item):
        return item in self.ports

    def __getitem__(self, item):
        if item in self.ports:
            return self.ports[item]
        elif item == 'name':
            return self.name
        else:
            return None

    def __setitem__(self, key, value):
        self.nw.send(self.gateway+':'+self.name, 'set', key, json.dumps(value))

    def set(self, key,value):
        self.ports[key] = value

    def __getattr__(self, item):
        return self.__getitem__(item)

    def __setattr__(self, key, value):
        if key in ['nw', 'name', 'gateway', 'ports']:
            super(Node, self).__setattr__(key, value)
        else:
            if key.startswith('on_'):
                self.__dict__[key] = value
                #super(Node, self).__setattr__(key, value) # todo: check if this is a method
            else:
                self.__setitem__(key, value)

class control:
    def __init__(self, nodename='control', inputs='', outputs='', handler = None):
        self.nw = NodeWire(nodename, process=self.process)
        self.nw.debug = False
        self.nodes = []
        self.sender = None
        self.inputs =  [{'port':i.split(':')[0] if ':' in i else i, 'props': i.split(':')[1] if ':' in i else 'Status', 'value':0} for i in  inputs.split()]
        self.outputs = [{'port':o.split(':')[0] if ':' in o else o, 'props': o.split(':')[1] if ':' in o else 'Status'} for o in outputs.split()]

        if not handler is None:
            for method in inspect.getmembers(handler, predicate=inspect.ismethod):
                if method[0].startswith('get_'):
                    port = method[0][4:]
                    ports = [p for p in self.outputs if p['port'] == port]
                    if ports != []: ports[0]['get'] = method[1]
                elif method[0].startswith('on_'):
                    port = method[0][3:]
                    ports = [p for p in self.inputs if p['port'] == port]
                    if ports != []: ports[0]['on'] = method[1]

    def get(self, item):
        ports = [p for p in self.outputs if p['port'] == item]
        if ports != []:
            port = ports[0]
            if 'get' in port:
                return port['get'](Sender=self.sender)
            elif 'value' in port:
                return port['value']
            else:
                return 0
        else:
            print('invalid port or attribute: {}'.format(item))
            return None
            # raise Exception('invalid port or attribute: {}'.format(item))

    def set(self, key, value):
        ports = [p for p in self.inputs if p['port'] == key]
        if ports != []:
            ports[0]['value'] = value
            if 'on' in ports[0]:
                if inspect.iscoroutinefunction(ports[0]['on']):
                    asyncio.create_task(ports[0]['on'](Sender=self.sender, Value=value))
                else:
                    ports[0]['on'](Sender=self.sender, Value=value)
        else:
            ports = [p for p in self.outputs if p['port'] == key]
            if ports != []:
                self.nw.send('re', 'portvalue', key, json.dumps(value))

    def __getattr__(self, item):
        if item in ['nw', 'nodes', 'sender', 'inputs', 'outputs']:
            return super(control, self).__getattr__(item)
        ports = [p for p in self.inputs if p['port']==item]
        if ports != []:
            port = ports[0]
            return port['value']
        else:
            return self.get(item)

    def __setattr__(self, key, value):
        if key in ['nw', 'nodes', 'sender', 'inputs', 'outputs']:
            super(control, self).__setattr__(key, value)
        else:
            if key.startswith('on_'):
                port = key[3:]
                ports = [p for p in self.inputs if p['port'] == port]
                if ports!=[]: ports[0]['on']=value
            elif key.startswith('get_'):
                port = key[4:]
                ports = [p for p in self.outputs if p['port'] == port]
                if ports != []: ports[0]['get'] = value
            else:
                self.set(key, value)

    def create_node(self, nodename, instance=None, got_it=None):
        if instance==None: instance = self.nw.gateway
        nodes = [n for n in self.nodes if n.name==nodename]
        if len(nodes) == 0:
            self.nw.send('cp','subscribe', nodename if instance==None else instance +':'+nodename, 'portvalue')
            self.nw.send('cp', 'getnode', nodename)
            if got_it!=None:
                will_get[nodename] = got_it
        else:
            if got_it!=None: got_it(nodes[0])

    def process(self, msg):
        if msg.Command == 'get':
            if msg.Port == 'ports':
                ports = ' '.join([o['port'] for o in self.outputs]) + ' ' + ' '.join([i['port'] for i in self.inputs])
                self.nw.send(msg.Sender, 'ports', ports)
            elif msg.Port == 'properties':
                ports = [p for p in self.inputs if p['port'] == msg.Params[1]]
                if ports != []:
                    self.nw.send(msg.Sender, 'properties', msg.Params[1], ports[0]['props'])
                else:
                    ports = [p for p in self.outputs if p['port'] == msg.Params[1]]
                    if ports != []:
                        self.nw.send(msg.Sender, 'properties', msg.Params[1], ports[0]['props'])
            else:
                self.sender = msg.Sender
                result = getattr(self, msg.Port, None)
                if inspect.iscoroutine(result):
                    def done(t):
                        self.nw.send(msg.Sender, 'portvalue', msg.Port, json.dumps(task.result()))
                        self.sender = None
                    task = asyncio.create_task(result)
                    task.add_done_callback(done)
                else:
                    self.nw.send(msg.Sender, 'portvalue', msg.Port, json.dumps(result))
                    self.sender = None
        elif msg.Command == 'set':
            if msg.Port in [p ['port'] for p in self.inputs]:
                self.sender = msg.Sender
                setattr(self, msg.Port, msg.Value)
                self.sender = None
                self.nw.send(msg.Sender, 'portvalue', msg.Port, json.dumps(getattr(self, msg.Port, None)))
        elif msg.Command == 'portvalue':
            if ':' in msg.Sender:
                sender = msg.Sender.split(':')
                msg.Sender = sender[1]
                instance = sender[0]
            else:
                instance = None
            senders = [s for s in self.nodes if s.name==msg.Sender]
            if senders!=[]:
                senders[0].set(msg.Port, msg.Value)
                if 'on_' + msg.Port in senders[0].__dict__:
                    senders[0].__dict__['on_' + msg.Port](msg.Sender)
        elif msg.Command == 'node':
            msg.Params[0] = msg.Params[0].replace("'", '"')
            msg.Params[0] = msg.Params[0].replace('None', 'null')
            msg.Params[0] = msg.Params[0].replace('True', 'true')
            msg.Params[0] = msg.Params[0].replace('False', 'false')
            nodevalue = json.loads(msg.Params[0])
            nodename = msg.Params[1]
            gateway = msg.Params[2]
            n = Node(self.nw, nodename, gateway)
            if n not in self.nodes:
                for key in nodevalue:
                    n.set(key, nodevalue[key])
                self.nodes.append(n)
            if nodename in will_get:
                will_get[nodename](n)
                del will_get[nodename]


if __name__ == '__main__':
    class Handler():
        def __init__(self):
            self.auto = False
            self.times = 0

        def lost_power(self, node):
            if sco.mains == 0 and self.auto: sco.ignition = 1
            self.times+=1

        def on_auto_switched(self, sender, value):
            self.auto = value

        def get_service_required(self, Sender):
            return self.times>10

        def connected(self, node=None):
            global sco
            if node is None:
                ctrl.create_node('sco', got_it=self.connected)
            else:
                sco = node
                sco.on_mains = self.lost_power

    ## MAIN PROGRAM
    handler = Handler()
    ctrl = control(inputs = 'auto_switch', outputs = 'service_required', handler=handler)
    ctrl.nw.on_connected = handler.connected
    ctrl.nw.debug = True
    ctrl.nw.run()